/*
 program: cronrgf.cmd
 type:    REXXSAA-OS/2, OS/2 2.x
 purpose: Unix-like cron; allow to repeatedly execute commands at given date/time
          defined in a control file
 version: 2.0.5 SHL
 date:    1992-06-10
 changed: 1993-03-19, bug fixed (caused no scheduling between 23:00 and 23:59),
                           logging added (so you can check for yourself)
         1993-11-05, added option to reread (/R) the crontab file at specified times (default every 60 minutes),
                           suggested by fmerrow@nyx10.cs.du.edu (Frank Merrow);
                     added option to execute all programs once since midnight (/M) and the time cronrgf was invoked,
                           suggested by sktoni@uta.fi (Tommi Nieminen);
                     added colorized output and option (/B) to use black and white
                           instead of ANSI-colors on output
         1993-11-22, added ability to look for crontab file in the directory in which cronrgf.cmd is stored,
                           if no path was explicitly given; hence one can start cronrgf.cmd from
                           any drive/directory without the need to specify a full path for the crontab file
         1994-04-04, changed the way the log-file is handled (will be closed after it was written to, hence one can
                           delete the log-file after a certain while, if one feels it has become too large)
         1996-04-30, changed testing of valid date from flag "M" to "C", in order to prohibit an error message,
                           if an illegal date was produced (needs datergf.cmd version 1.6 for this to be true)
         1997-05-03, changed a bug in the dispatch-loop (thanks to <Del_Brown@afcc.fabrik.com> !)

         2010-09-03 SHL Rework command quoting to avoid start errors
         2010-09-03 SHL Rework error handler
         2012-08-06 SHL Die if datergf.cmd not in PATH
         2012-08-06 SHL Disable colorization if ScrColor.cmd not available
         2014-07-13 SHL Support range/step syntax
         2014-07-13 SHL Sanitize function name case
         2014-07-13 SHL Clean up code layout
         2014-07-13 SHL Clean up args parsing - needs more work
         2014-07-13 SHL Version 2.0.5
         2016-03-30 SHL Use Ctrl-C to request crontab re-read, 2 Ctrl-C to stop
         2016-03-30 SHL Use IsoDateTime() for dates in logs
         2016-03-30 SHL Version 2.0.6
         2016-04-08 SHL Rework case to my style - this is going to be a fork rather than just an update
         2016-05-30 SHL Make SkelRexx Globals compatible
         2016-05-30 SHL Rework command execution error handling - log and continue
         2016-05-30 SHL Version 2.0.7 beta
         2016-07-14 SHL Version 2.0.7

 author:  Rony G. Flatscher
          Rony.Flatscher@wu-wien.ac.at

 needs:   DATERGF.CMD, SCRCOLOR.CMD, some RxFuncts (loaded automatically)

 usage:   CRONRGF [/T] [/M] [/R[nn]] [/B] crontab

    Reads crontab file and executes command[s] repeatedly according to schedule.
    The crontab file has to be formatted like in Unix; unlike the Unix-version a '%' is
    treated like any other character; empty lines and ones starting with
    a semi-colon (;) or a (#) are ignored.

    Switch '/T[estmode]': the crontab file is read and the user is presented with the date/times
          and commands to be executed upon them. The planned and truly scheduled
          times are written to the log-file.

    Switch '/M[idnight]': all commands which were scheduled the same day, starting from
          midnight to the time of first invocation are executed once.

    Switch '/R[nn]': check crontab file at least every 'nn' (default 60) minutes whether it changed; if so, reread
          it immediately and set the new schedule-times. If set, the crontab file will be checked for changes after
          executing commands

    Switch '/B[lackAndWhite]': do not colorize output strings (e.g. for usage in more, less etc.).
          This switch merely suppresses the ANSI-color-sequences attached to the
          strings.

    example:  CRONRGF /TEST testcron
              execute statements in file 'testcron' in testmode

    example for a control-file:

        ; Sample file for CRONRGF.CMD
        ;
        ; This is a comment (starts with a semicolumn)
        # This is a comment too (starts with the Unix-comment symbol)
        ; empty lines are ignored too...

        ; LAYOUT OF THE CRONTAB FILE:
        ;    * * * * * command
        ; or
        ;    minute hour day month weekday command
        ; where minute  ranges from 0-59,
        ;       hour    ranges from 0-23,
        ;       day     ranges from 1-31,
        ;       month   ranges from 1-12,
        ;       weekday ranges from 1-7 (1 = Monday, 2 = Tuesday, ..., 7 = Sunday)
        ;
        ; you can give a list of values, separated by a comma (,), e.g. "1,3,7"
        ; you can give a range of values, separated by a dash (-), e.g. "1-5"
        ; you can give a star (*) instead of a value, meaning entire range of all valid values
        ; * and ranges can be suffixed by / and a step count
        ;
        ; the given command is only executed when all criteriae are fullfilled !
        ;
        ; restriction: unlike to Unix, the percent-sign (%) is treated like any other character and
        ;              not as a new-line
        ;

        # the following command "@ECHO HI, I am Nr. 1 to be echoed every minute" would be
        # executed every minute
        *  *  *  *  *  @ECHO Hi, I am Nr. 1 to be echoed every minute & pause

        59 23 31 12 5 command, one minute before year's end, and only if the last day is a Friday

        ; comment: every year at 17:45 on June 7th:
        45 17  7  6  *  dir c:\*.exe

        ; comment: on every full quarter of an hour
        ;          at midnight, 6 in the morning, noon, 6 in the evening
        ;          on the 1st, 15th and 31st of
        ;          every month on
        0,15,30,45   0,6,12,18   1,15,31   *   *   backup c:\*.* d:\ /s

        ; at noon on every day, if it is a weekday (Mo-Fri):
        0 12 * * 1-5 XCOPY Q:\* D:\ /s

        ; every minute in January, March, May, July, September and November:
        *  *  *  1,3,5,7,9,11  *  dir c:\*.cmd

        # at the last day of the year at 23:01, 23:02, 23:03, 23:05, 23:20, 23:21,
        # 23:22, 23:23, 23:24, 23:25, 23:30, 23:31, 23:32, 23:33, 23:34, 23:35,
        # 23:59
        1,2,3,5,20-25,30-35,59   23   31   12   *   COPY D:\*.log E:\backup

        ; make backups of OS2.INI and OS2SYS.INI on every first monday of a month,
        ; at 9 o'clock in the morning
        0 9 1-7 * 1 showini /bt d:\os2\os2.ini
        0 9 1-7 * 1 showini /bt d:\os2\os2sys.ini

        ; at midnight on every month:
        0 0 1 * * tapebackup /all

        ; execute every minute, no restrictions:
        *  *  *  *  *  @ECHO Hi, I am Nr. 2 to be echoed every minute & pause

        # execute every minute in January, February, March only !
        * *  *  1,2,3  *  any-command any-arguments

        # execute every day at midnight
        0 0 * * * any-command any-arguments

        # execute every wednesday at midnigth !
        0 0 * * 3 any-command any-arguments

        # this is a comment which concludes the sample file ===========================

All rights reserved, copyrighted 1992, no guarantee that it works without
errors, etc. etc.

Donated to the public domain granted that you are not charging anything (money
etc.) for it and derivates based upon it, as you did not write it,
etc. if that holds you may bundle it with commercial programs too.

You may freely distribute this program, granted that no changes are made
to it.

Please, if you find an error, post me a message describing it, I will
try to fix and rerelease it to the net.

*/

signal on Error name ErrorVLX
signal on FAILURE name ErrorVLX
signal on NOTREADY name ErrorVLX
signal on NOVALUE name ErrorVLX
signal on SYNTAX name ErrorVLX
call on Halt

gVersion = '2.0.7'

Globals = 'gArgList. gCmdName gDbgLvl gEnv gFileName gTesting gTmpDir gVerbose gVersion'

/* Ensure RexxUtil functions are loaded */
if RxFuncQuery('SysLoadFuncs') then do
   call RxFuncAdd 'SysLoadFuncs', 'RexxUtil', 'SysLoadFuncs'
   call SysLoadFuncs
end

if SysSearchPath('PATH', 'datergf.cmd') == '' then do
   say 'Can not find required datergf.cmd script in PATH'
   exit -1
end

Main:

global. = ''                    /* default colors off etc. */
global.eTestmode = 0            /* default: no testmode */
global.eFirstInvocation = 0     /* don't execute commands since midnight up to invocation time */
global.eCheckCronFile = 0       /* don't check whether crontab file was changed */

stemSchedule. = ''              /* default for empty array elements */

cmdline = arg(1)

pos1 = pos('/B', translate(cmdline))

if pos1 > 0 then do
  /* no screen-colors */
  /* Strip /B from argument list */
  /* Find next space or slash */
  pos2 = verify(cmdline, '/ ', 'Nomatch', pos1 + 2)
  if pos2 = 0 then
    cmdline = substr(cmdline, 1, pos1 - 1)
  else
    cmdline = substr(cmdline, 1, pos1 - 1) strip(substr(cmdline, pos2 - 1))
end
else do
  /* Set up colors if ScrColor available */
  if SysSearchPath('PATH', 'ScrColor.cmd') == '' then
    say 'ScrColor.cmd not found in PATH - using black and white output'
  else do
    /* get screen-colors */
    parse value ScrColor() with global.eScrNorm    global.eScrInv,
                                global.eTxtNorm    global.eTxtInf,
                                global.eTxtHi      global.eTxtAla,
                                global.eTxtNormInv global.eTxtInfInv,
                                global.eTxtHiInv   global.eTxtAlaInv .
  end
end

if cmdline = '' then
  signal Usage

/* have three arguments ? */
parse var cmdline '/'switch1 '/'switch2 '/'switch3 filein

/* have two arguments ? */
if filein = '' then
  parse var cmdline '/'switch1 '/'switch2 filein

/* have one argument ? */
if filein = '' then
  parse var cmdline '/'switch1 filein

/* up to three arguments reversed ? */
if filein = '' then
  parse var cmdline filein '/'switch1 '/'switch2  '/'switch3

if filein = '' & switch1 = '' then
  filein = cmdline

if filein = '' | filein = '?' then
  signal Usage

switches = translate(left(switch1, 1) || left(switch2, 1) || left(switch3, 1))

/* check whether switches are valid */
if verify(switches, 'TMR ') <> 0 then do
  call say_c global.eTxtAla || 'CRONRGF: unknown switch in [' || global.eTxtHi || cmdline || global.eTxtAla || '].'
  call say_c
  call Beep 2500, 100
  signal Usage                          /* wrong switch */
end

switch_text = ''

if pos('T', switches) > 0 then do
  global.eTestmode = 1
  switch_text = switch_text '/T'
end

if pos('M', switches) > 0 then do
  global.eFirstInvocation = 1
  switch_text = switch_text '/M'
end

pos = pos('R', switches)
if pos > 0 then do
  select
  when pos = 1 then tmp = switch1
  when pos = 2 then tmp = switch2
  otherwise tmp = switch3
  end

  parse upper var tmp 'R' minutes

  if minutes = '' then
    minutes = 60                        /* default to 60 minutes */
  else if \ datatype(minutes, 'N') | minutes < 5 | minutes > 43200 then do
    call say_c  global.eTxtAla || 'CRONRGF: wrong minute-value in switch [/' || global.eTxtHi || tmp || global.eTxtAla || '].'
    call say_c  global.eTxtAla || '         (valid range 5-43200 minutes = 5 minutes to 30 days).'
    call say_c
    call Beep 2500, 100
    signal Usage                        /* wrong switch */
  end
  switch_text = switch_text '/R' || minutes
  global.eMinutesToSleep = minutes
  global.eSecondsToSleep = minutes * 60
  global.eCheckCronFile = 1
end /* if /R */

/* construct log file name */
parse source . . full_path_of_this_procedure
global.eLogFile = substr(full_path_of_this_procedure, 1,,
                         lastpos('.', full_path_of_this_procedure)) || 'LOG'

filein = strip(filein)          /* get rid of leading and trailing spaces */
global.eFilein = stream(filein, 'C', 'QUERY EXISTS')

if global.eFilein = '' then do
  /* supply drive & path of cronrgf.cmd, if not supplied by user */
  if filespec('Drive', filein) = '' & filespec('Path', filein) = '' then do
    parse source . . this
    filein = filespec('Drive', this) || filespec('Path', this) || filein
    global.eFilein = stream(filein, 'C', 'QUERY EXISTS')
  end

  if global.eFilein = '' then
    call Stop_It 'crontab file [' || filein || '] does not exist.'
end

call Write_log left('', 80, '-')
call Write_log '(Logging-session started:' IsoDateTime() || ')'
call Write_log '  (Switches in effect) [' || strip(switch_text) || ']'

call read_cronfile                      /* read given crontab file */

if global.0 > 0 then
  call dispatch                         /* start dispatching - does not return */

call Stop_It 'Nothing to schedule. Program ended.'

/**
 * Dispatch commands forever or until killed
 */

dispatch: procedure expose global. stemSchedule.

  if global.eCheckCronFile then
    sleeping_time = global.eSecondsToSleep
  else                        /* take care of DosSleep()-unsigned long; to be safe just sleep a maximum */
    sleeping_time = 2678400  /* of 31 days (= 31 * 24 * 60 * 60  == 2.678.400 seconds) at a time       */

  gHaltTime = 0
  gWantReread = 0

  do forever_loop = 1 to 1 by 0

    call say_c right('', 79, '=')
    call say_c
    call say_c 'Scheduling commands given in crontab file:' /* show user which file is being used */
    call say_c
    call say_c '  [' || global.eTxtHi || global.eFilein || global.eTxtInf || ']'

    if global.eCheckCronFile then
      call say_c '  (being checked after executing commands, at least every'  global.eTxtHi || global.eMinutesToSleep  global.eTxtInf || 'minutes)'

    if global.eFirstInvocation then
      call say_c '  (executing once all commands between midnight and now, switch:' global.eTxtHi ||  '/M' || global.eTxtInf || ')'

    call say_c

    call Schedule_Next

    /* show user which command(s) will be executed when */
    parse var stemSchedule.1 1 next_date_time 18 /* get next date/time */

    /* get actual date/time */
    act_date_time = date('S') time()

    if \ global.eTestmode then
      call say_c 'command[s] being scheduled on:'  global.eTxtHi || IsoDateTime(next_date_time)
    else do
      /* Test mode */
      if \ global.eFirstInvocation then
        act_date_time = next_date_time
      call say_c 'command[s] being scheduled on:'  global.eTxtHi || IsoDateTime(act_date_time)
    end

    call say_c

    /* Show items selected to run */
    do i = 1 to stemSchedule.0
      parse var stemSchedule.i 1 tmp_date_time 18 status index
      if tmp_date_time > next_date_time then
        leave

      if status = 'OK' then do
        call say_c '  [' || global.eTxtHi || value('global.' || index || '.eCommand.eValues') || global.eTxtInf || ']'
        if global.eTestmode then do
          /* Write info to log file */
          call Write_log '  (Testmode)' act_date_time '(scheduled for',
                          next_date_time || ') [' ||,
                          value('global.' || index || '.eCommand.eValues') || ']'
        end
      end
    end /* do */

    call say_c

    /* Calc time until next scheduled task */
    difference = datergf(next_date_time, '-S', act_date_time)

    if difference > 0 then
      seconds_to_sleep = datergf(difference, 'SEC') % 1
    else
      seconds_to_sleep = 0

    if global.eTestmode then do
      /* Test mode */
      if \ global.eFirstInvocation then
        act_date_time = next_date_time

      call say_c right('', 79, '=')
      call say_c 'Testmode (dispatch):    next_invocation ='  global.eTxtHi || IsoDateTime(next_date_time)
      call say_c 'Testmode (dispatch):   actual date/time ='  global.eTxtHi || IsoDateTime(act_date_time)
      if difference < 0 then
        call say_c 'Testmode (dispatch):                     ' global.eTxtHi || 'Immediate !'
      else
        call say_c 'Testmode (dispatch): difference in days =' global.eTxtHi || difference,
                   global.eTxtInf || '=' global.eTxtHi || seconds_to_sleep  global.eTxtInf || 'seconds'

      call say_c right('', 79, '=')

      call say_c 'Testmode (dispatch):'
      call say_c '   input:'
      do i = 1 to global.0
        call say_c '     [' || global.eTxtHi || right(i, length(global.0)) || global.eTxtInf || ']' global.original.i
      end
      call say_c right('', 79, '=')

      call say_c '   schedule list:'
      do i = 1 to stemSchedule.0
        parse var stemSchedule.i . . . gIndex
        call say_c '    ' stemSchedule.i '[' || global.eTxtHi || value('global.' || gIndex || '.eCommand.eValues') || global.eTxtInf || ']'
      end
      call say_c right('', 79, '=')

      call say_c 'Press any key to see next scheduled action or "q" to quit.'
      call Beep 500, 100
      if translate(SysGetKey('NOECHO')) = 'Q' then
        signal halt
    end /* testmode */

    else do
      /* Live mode */
      do while seconds_to_sleep > 0
        sleeping         = min(seconds_to_sleep, sleeping_time)
        seconds_to_sleep = max(0, seconds_to_sleep - sleeping_time)

        call SysSleep sleeping          /* sleep */

        /* update sleeping-time, get actual date/time */
        difference = datergf(next_date_time, '-S', date('S') time())

        if difference > 0 then
          seconds_to_sleep = datergf(difference, 'SEC') % 1
        else
          seconds_to_sleep = 0

        /* if crontab file check pending and another sleeping interval is coming up, then reread the crontab file */
        if global.eCheckCronFile & seconds_to_sleep > 0 then do
          if cronfile_changed() then do
            call do_the_cronfile_reread
            iterate forever_loop        /* reschedule */
          end
        end

        if gWantReread & seconds_to_sleep > 0 then do
          gWantReread = 0
          call do_the_cronfile_reread
          iterate forever_loop        /* reschedule */
        end

      end /* while need to sleep */

      act_date_time = date('S') time()  /* now get the up-to-date time, was bug pointed to by Del Brown */

      /* Run scheduled commands */
      do i = 1 to stemSchedule.0
        parse var stemSchedule.i 1 date_time 18 status index
        if date_time > act_date_time then
          leave

        if status = 'OK' then do
          /* Run command in a minimized session which closes automatically after the
             command finishes executing
           */
          commandString = value('global.' || index || '.eCommand.eValues')
          if 0 then trace '?A' ; nop
          /* 2010-09-03 SHL drop extra quotes to avoid start errors */
          title = '"CRONRGF:' IsoDateTime(date_time) space(translate(commandString, ' ', '"')) || '"'
          /* 2010-09-03 SHL do not quote command string - it really confuses shells */
	  call on Error name ErrorVLX
	  drop gErrCondition
          address CMD '@start' title '/c /win /min /b' commandString
	  signal on Error name ErrorVLX
	  call Write_log '(dispatched)' IsoDateTime() '(scheduled for' date_time || ') [' || commandString || ']'
	  if symbol('gErrCondition') == 'VAR' then
	    call Write_log '(***ERROR)' IsoDateTime() '(failed with error' gErrCondition || ')'

	end /* if OK */
      end /* stemSchedule */

    end /* live mode */

    global.eFirstInvocation = 0 /* programs from midnight up to invocation time were executed, stop that behavior */

    /* Reread crontab file if requested and if crontab file changed or in testmode */
    if global.eCheckCronFile & (cronfile_changed() | global.eTestmode) then do
      call do_the_cronfile_reread
      iterate forever_loop              /* reschedule */
    end

    /* Change the status of the executed programs to NOK */
    do i = 1 to stemSchedule.0
      parse var stemSchedule.i 1 date_time 18 status index
      if date_time > act_date_time then
	leave
      stemSchedule.i = act_date_time 'NOK' index
    end

  end /* do forever_loop */

  return				/* Should never get here */

/**
 * Process schedule and calculate scheduled times sorted ascending order
 */

Schedule_Next: procedure expose global. stemSchedule.

  /* as long as no viable date/time to schedule was found, iterate */
  main_run = 0

  do while word(stemSchedule.1, 3) <> 'OK'

    main_run = main_run + 1             /* count loops until a valid day was found for any of the commands */

    if global.eTestmode then
     call say_c 'Testmode (scheduling): main loop ='  global.eTxtHi || main_run

    if main_run > 50 then
     call Stop_It 'Schedule_Next(): aborting after 50 attempts to produce a valid date!'

    do i = 1 to stemSchedule.0
      parse var stemSchedule.i 1 year 5 month 7 day 9 10 hour 12 13 minute 15 18 disp_status glob_index

      if disp_status = 'OK' then
	iterate 			/* Already scheduled */

      old_year  = year
      old_month = month
      old_day   = day
      old_hour  = hour

      /* defaults */
      first_minute = word(global.glob_index.eMinute.eValues, 1)
      first_hour   = word(global.glob_index.eHOur.eValues, 1)
      first_day    = word(global.glob_index.eDay.eValues, 1)
      first_month  = word(global.glob_index.eMonth.eValues, 1)

      /* minute */
      do j = 1 to global.glob_index.eMinute.0
	tmp = word(global.glob_index.eMinute.eValues, j)
	if tmp > minute then
	  leave
      end

      if j > global.glob_index.eMinute.0 then /* minutes to wrap around */
	hour   = hour + 1
      else                              /* minutes within same hour */
	minute = tmp

      /* hour */
      do j = 1 to global.glob_index.eHour.0
	tmp = word(global.glob_index.eHour.eValues, j)
	if tmp >= hour then
	  leave
      end
      if j > global.glob_index.eHour.0 then /* hours to wrap around */
	day    = day + 1
      else                              /* hours within same day */
	hour   = tmp

      ok = 'NOK'                        /* default: no date found yet */
      run = 0
      do 50                             /* try 50 times to produce a valid date */
	run = run + 1
	/* day */
	do j = 1 to global.glob_index.eDay.0
	  tmp = word(global.glob_index.eDay.eValues, j)
	  if tmp >= day then leave
	end

	if j > global.glob_index.eDay.0 then do
	  /* days to wrap around */
	  day    = first_day
	  month  = month + 1
	end
	else                            /* days within same month */
	  day    = tmp

	/* month */
	do j = 1 to global.glob_index.eMonth.0
	  tmp = word(global.glob_index.eMonth.eValues, j)
	  if tmp >= month then leave
	end

	if j > global.glob_index.eMonth.0 then do
	  /* months to wrap around */
	  day    = first_day
	  month  = first_month
	  year   = year + 1
	end
	else do
	  /* months within same year */
	  if month <> tmp then          /* did the month change ? */
	    day = first_day
	  month  = tmp
	end

	select
	  when old_year < year | old_month < month | old_day < day then
	    next_invocation = year || month || day first_hour || ':' || first_minute || ':00'
	  when old_hour < hour then
	    next_invocation = year || month || day       hour || ':' || first_minute || ':00'
	  otherwise
	    next_invocation = year || month || day       hour || ':' ||       minute || ':00'
	end

	if global.eTestmode then
	  call say_c 'Testmode (scheduling): next_invocation ='  global.eTxtHi || next_invocation datergf(next_invocation, 'DN')

	/* check whether day-of-week is o.k. */
	if datergf(next_invocation, 'C') = '' then do
	  /* illegal date produced, e.g. 19950231 ? */
	  day   = first_day
	  month = month + 1
	  if month > 12 then do
	    month = first_month
	    year  = year + 1
	  end
	  iterate
	end

	next_Weekday = datergf(next_invocation, 'DI') /* get weekday */

	/* using pos because weekdays are in the form of 01, 02, ..., 07 */
	if pos(next_Weekday, global.glob_index.eWeekday.eValues) = 0 then do
	  /* invalid weekday ? */
	  next_invocation = datergf(next_invocation, '+', '1') /* add one day to present date */
	  parse var next_invocation 1 year 5 month 7 day 9
	  iterate
	end

	ok = ' OK'                      /* o.k. to invoke, because valid date */
	leave
      end /* do 50 */

      if global.eTestmode then do
	call say_c 'Testmode (scheduling): date/time-loop ='  global.eTxtHi || run  global.eTxtInf || 'time[s]'
	call say_c
      end

      /* schedule list format:
	 DATE TIME STATUS INDEX-INTO-GLOBAL-ARRAY
      */
      stemSchedule.i = next_invocation ok glob_index

    end /* do i */

    call sort_schedule_list

  end /* do while item unscheduled */

  return

/**
 * sort schedule list in ascending order
 */

sort_schedule_list: procedure expose global. stemSchedule.

  length = 21          /* length of substr to compare, includes status */
  /* calculate M - number of stem sort passes */
  M = 1
  do while (9 * M + 4) < stemSchedule.0
    M = M * 3 + 1
  end

  /* sort stem */
  do while M > 0
    K = stemSchedule.0 - M
    do J = 1 to K
      Q = J
      do while Q > 0
	L = Q + M
	if substr(stemSchedule.Q, 1, length) <= substr(stemSchedule.L, 1, length) then leave
	/* switch elements */
	tmp            = stemSchedule.Q
	stemSchedule.Q = stemSchedule.L
	stemSchedule.L = tmp
	Q = Q - M
      end
    end
    M = M % 3
  end /* do while */

  return

/**
 * validate crontab line
 */

check_it_out: procedure expose global. stemSchedule.

  crontab_line = arg(1)
  line_no  = arg(2)

  parse var crontab_line sMinute sHour sDay sMonth sWeekday sCommand
  /* 2010-09-03 SHL strip leading whitespace left by parse */
  sCommand = strip(sCommand)

  line_no = setup_minutes(sMinute, line_no) /* setup minute-values */

  if line_no <> 0 then                  /* setup hour-values */
    line_no = setup_hours(sHour, line_no)

  if line_no <> 0 then                  /* setup day-values */
    line_no = setup_days(sDay, line_no)

  if line_no <> 0 then                  /* setup month-values */
    line_no = setup_months(sMonth, line_no)

  if line_no <> 0 then                  /* setup weekday-values */
    line_no = setup_weekdays(sWeekday, line_no)

  if line_no <> 0 then do
    /* setup command-values */
    global.line_no.eCommand.0 = 1
    global.line_no.eCommand.eValues = sCommand
  end

  return line_no

/**
 * parse and setup minutes
 * arg(1) - minute string
 * arg(2) - index into global array
 * return index or 0 if error
 */

setup_minutes: procedure expose global. stemSchedule.
  sMinute = arg(1)
  iIndex = arg(2)
  default.0 = 60
  default.values = '00 01 02 03 04 05 06 07 08 09 10 11 12 13 14 15 16 17 18 19',
		   '20 21 22 23 24 25 26 27 28 29 30 31 32 33 34 35 36 37 38 39',
		   '40 41 42 43 44 45 46 47 48 49 50 51 52 53 54 55 56 57 58 59'

  string = parse_it('eMinute', sMinute, iIndex, 0, 59, default.0, default.values)

  if string <> '' then do
    call say_c  global.eTxtAla || 'CRONRGF: error in minute-format' global.eTxtHi || '[' || string || ']'
    call say_c
    return 0
  end

  return iIndex

/**
 * parse and setup hours
 * arg(1) - hour string
 * arg(2) - index into global array
 * return index or 0 if error
 */

setup_hours: procedure expose global. stemSchedule.
  sHour = arg(1)
  iIndex = arg(2)
  default.0 = 24
  default.values = '00 01 02 03 04 05 06 07 08 09 10 11 12 13 14 15 16 17 18 19',
		   '20 21 22 23'

  string = parse_it('eHour', sHour, iIndex, 0, 23, default.0, default.values)

  if string <> '' then do
    call say_c  global.eTxtAla || 'CRONRGF: error in hour-format' global.eTxtHi || '[' || string || ']'
    call say_c
    return 0
  end

  return iIndex

/**
 * parse and setup days
 * arg(1) - day string
 * arg(2) - index into global array
 * return index or 0 if error
 */

setup_days: procedure expose global. stemSchedule.

  sDay = arg(1)
  iIndex = arg(2)
  default.0 = 31
  default.values = '01 02 03 04 05 06 07 08 09 10 11 12 13 14 15 16 17 18 19',
		   '20 21 22 23 24 25 26 27 28 29 30 31'

  string = parse_it('eDay', sDay, iIndex, 1, 31, default.0, default.values)

  if string <> '' then do
    call say_c  global.eTxtAla || 'CRONRGF: error in day-format' global.eTxtHi || '[' || string || ']'
    call say_c
    return 0
  end

  return iIndex

/**
 * parse and setup months
 * arg(1) - month string
 * arg(2) - index into global array
 * return index or 0 if error
 */

setup_months: procedure expose global. stemSchedule.

  sMonth = arg(1)
  iIndex = arg(2)
  default.0 = 12
  default.values = '01 02 03 04 05 06 07 08 09 10 11 12'

  string = parse_it('eMonth', sMonth, iIndex, 1, 12, default.0, default.values)

  if string <> '' then do
    call say_c  global.eTxtAla || 'CRONRGF: error in month-format' global.eTxtHi || '[' || string || ']'
    call say_c
    return 0
  end

  return iIndex

/**
 * parse and setup weekdays
 * arg(1) - weekday string
 * arg(2) - index into global array
 * return index or 0 if error
 */

setup_weekdays: procedure expose global. stemSchedule.

  sWeekday = arg(1)
  iIndex = arg(2)
  default.0 = 7
  default.values = '01 02 03 04 05 06 07'

  string = parse_it('eWeekday', sWeekday, iIndex, 1, 7, default.0, default.values)

  if string <> '' then do
    call say_c  global.eTxtAla || 'CRONRGF: error in weekday-format' global.eTxtHi || '[' || string || ']'
    call say_c
    return 0
  end

  return iIndex

/**
 * parse values list and set eValues string in array
 * arg(1) = eName ("element name" in array)
 * arg(2) = values string containing comma sep list of items
 * arg(3) = index into global array
 * arg(4) = lower bound (inclusive)
 * arg(5) = upper bound (inclusive)
 * arg(6) = default number of elements
 * arg(7) = default values
 * return error message text or empty string if OK

 * values string items are of the form
 *   *
 *   * / step (no spaces)
 *   comma sep list of
 *     number
 *     first-last/step
 */

parse_it: procedure expose global. stemSchedule.

  eName          = arg(1)
  sValues        = arg(2)               /* Assume already stripped and no interior spaces */
  iIndex         = arg(3)
  lower          = arg(4)
  upper          = arg(5)
  default.0      = arg(6)               /* Number of items */
  default.values = arg(7)               /* Item string */

  stem = 'global.iIndex.' || eName || '.' /* build string of array-e-Name */
  lastValue = 0

  err_msg = ''
  curValue = sValues

  /* Process special cases */

  if curValue = '*' then do
    /* build all legal values */
    interpret(stem || '0 =' default.0)
    interpret(stem || 'eValues =' default.values)
    return ''
  end

  if left(curValue, 2) == '*/' then do
    parse var curValue '*/'step
    if \ datatype(step, 'N') then
      err_msg = '"' || step || '"' 'is not numeric'
    else if step <= 0 | step > default.0 then
      err_msg = step '>' default.0 'upper bound'
    else do
      s = ''
      c = 0
      do i = 1 to default.0 by step
	c = c + 1
	s = s word(default.values, i)
      end
      s = strip(s)
      interpret(stem || '0 =' c)
      interpret(stem || 'eValues =' s)
      return ''
    end
  end

  /* Process comma separated list */

  interpret(stem || '0 = 0')            /* set number of elements to 0 */
  interpret(stem || 'eValues = ""')     /* delete values */

  do while sValues <> '' & err_msg = ''

    if pos(',', sValues) > 0 then
      parse var sValues curValue','sValues /* get value from list */
    else do
      curValue = sValues                /* last item */
      sValues = ''
    end

    /* Have step? */
    if pos('/', curValue) = 0 then do
      range = curValue
      step = 1
    end
    else
      parse var curValue range'/'step

    /* range of values ? */
    if pos('-', range) > 0 then
      parse var range start '-' end
    else do
      /* single value */
      start = range
      end   = range
    end

    /* values OK? */
    err_msg = ''
    select
    when \ datatype(start, 'N') then err_msg = '"' || start || '"' 'is not numeric'
    when \ datatype(end,   'N') then err_msg = '"' || end || '"' 'is not numeric'
    when \ datatype(step,  'N') then err_msg = '"' || step || '"' 'is not numeric'
    when start < lastValue      then err_msg = start '<' lastValue 'lower bound'
    when start < lower          then err_msg = start '<' lower 'lower bound'
    when start > end            then err_msg = start '>' end
    when end   > upper          then err_msg = end   '>' upper 'upper bound'
    otherwise
      err_msg = ''
    end

    if err_msg \== '' then do
      /* Bad value - clean up and complain */
      interpret(stem || '0  = ""')      /* delete number of array elements */
      interpret(stem || 'eValues = ""') /* delete values */
      err_msg = err_msg '(part in error: "' || curValue || '")'
      return err_msg
    end

    /* Build space separated value string */
    do i = start to end by step
      interpret(stem || '0 = ' || stem || '0 + 1') /* increment number of elements */
      interpret(stem || 'eValues = ' || stem || 'eValues' right(i, 2, '0')) /* append next value */
    end

    lastValue = end + 1                 /* for error checking */

  end /* while sValues */

  if err_msg \== '' then do
    /* Bad value - clean up and complain */
    interpret(stem || '0  = ""')        /* delete number of array elements */
    interpret(stem || 'eValues = ""')   /* delete values */
    err_msg = err_msg '(part in error: "' || curValue || '")'
  end

  return err_msg

/**
 * reread crontab file
 */

do_the_cronfile_reread: procedure expose global. stemSchedule.

  call say_c
  call say_c 'CRONRGF: crontab file [' || global.eTxtHi || global.eFilein || global.eTxtInf || '] changed, rereading ...'

  if global.eTestmode then
    leadin = '  (Testmode)'
  else
    leadin = '            '

  call Write_log leadin IsoDateTime() 'crontab file [' || global.eFilein || '] changed, rereading ...'

  call read_cronfile                    /* read new crontab file */

  if global.0 = 0 then                  /* no valid entries found */
    call Stop_It 'crontab file [' || global.eFilein || '] not a crontab file anymore ! Aborting ...'

  return

/**
 * Read contents of crontab file
 */

read_cronfile: procedure expose global. stemSchedule.

  if global.eFirstInvocation then do
    /* execute once from midnight to now */
    date_time = datergf(date('S'), '-', datergf('1', 'SECR')) /* Subtract 1 second from 00:00am */
  end
  else
    date_time = date('S') time()

  call SysFileTree global.eFilein, 'fileList', 'F'
  if fileList.0 <> 1 then
    call Stop_It 'control-file [' || global.eFilein || '] does not exist anymore !'

  global.eFilein.eState = fileList.1       /* remember crontab file state */

  global.0 = 0
  drop stemSchedule.
  stemSchedule. = ''   /* Fill schedule with empty items */
  iSchedule = 0        /* Set number of items in schedule */
  line_no = 1          /* Set next line number */

  do while lines(global.eFilein) > 0

    line = linein(global.eFilein)

    /* no empty lines and no comments */
    s = left(strip(line), 1)
    if line = '' | s = ';' | s = '#' then iterate

    call say_c 'parsing ['  global.eTxtInf || line || global.eTxtInf || ']'

    if check_it_out(line, line_no) > 0 then do
      /* crontab line is valid, add to schedule */
      iSchedule = iSchedule + 1
      global.original.line_no = line
      stemSchedule.iSchedule =  date_time 'NOK' line_no /* NOK = not o.k., get next date/time */
      line_no = line_no + 1		/* next line# */
    end
  end /* do lines */
  call say_c

  call Write_log '  (crontab file)' IsoDateTime() '[' || strip(global.eFilein.eState) || ']'

  global.eOriginal.0 = line_no - 1      /* number of crontab lines scanned */
  global.0 = iSchedule                  /* set number of schedule items */
  stemSchedule.0 = iSchedule            /* set number of schedule items */
  call stream global.eFilein, 'C', 'CLOSE' /* make sure, file is closed */

  return

/**
 * check, if crontab file has changed
 * @returns 1 if changed else 0
 */

cronfile_changed: procedure expose global. stemSchedule.

  call SysFileTree global.eFilein, 'fileList', 'F'
  if fileList.0 <> 1 then
    call Stop_It 'crontab file [' || global.eFilein || '] does not exist anymore ! Aborting ...'

  return global.eFilein.eState <> fileList.1

/**
 * Handle ctrl-c and ctrl-break
 * If 2 hits in less than .25 seconds shut down otherwise reread crontab
 */

Halt:

  if condition('I') == 'CALL' then do
    haltTime = time('L')
    /* Convert hh:mm:ss.uuuuuu to number
			  string     otbl       itbl
			  HHMMSS.UU              HH:MM:SS.UUUUUU */
    haltTime = translate('ABDEGHIJK', haltTime, 'ABCDEFGHIJKLMNO')

    if haltTime - gHaltTime > .25 then do
      gHaltTime = haltTime
      gWantReread = 1
      return
    end
  end

  /* Does not return */
  call Stop_It 'User interrupted program.'

/**
 * Clean up and close open files
 */

Stop_It:

  if global.eLogFile <> '' then do
    if arg(1) <> '' then
      call Write_log '  (***ERROR:' IsoDateTIme() '-'  arg(1) || ')'

    call Write_log '(Logging-session ended:' IsoDateTime() || ')'

    if left(stream(global.eLogFile), 1) = 'R' then
      call stream global.eLogFile, 'C', 'CLOSE' /* close log-file */
  end

  if arg(1) <> '' then call say_c  global.eTxtAla || 'CRONRGF:'  global.eTxtAlaInv || arg(1)

  if global.eFilein <> '' then do
    if left(stream(global.eFilein), 1) = 'R' then
      call stream global.eFilein, 'C', 'CLOSE' /* make sure, file is closed */
  end

  exit -1

/**
 * Display usage help
 */

Usage:

  call say_c  global.eTxtHi || 'CRONRGF.CMD'  global.eTxtInf || '- Unix-like cron; executes commands in file repeatedly.'
  call say_c 'Version' gVersion
  call say_c
  call say_c 'Usage:'
  call say_c
  call say_c  global.eTxtHi || '      CRONRGF [/T] [/M] [/R[nn]] [/B] crontabfile'
  call say_c
  call say_c 'Reads file "' || global.eTxtHi || 'crontabfile' || global.eTxtInf || '" and executes command[s] repeatedly according to it.'
  call say_c "The crontab file has to be formatted like in Unix; unlike the Unix-version a '%' is"
  call say_c 'treated like any other character; empty lines and ones starting with'
  call say_c 'a semi-colon (;) or with hash (#) are ignored.'
  call say_c
  call say_c 'Switch "' || global.eTxtHi || '/T' || global.eTxtInf || '": crontabfile is read and the user is presented with the date/times'
  call say_c '       and commands to be executed upon them.'
  call say_c 'Switch '" || global.eTxtHi || '/M' || global.eTxtInf || "': execute once the programs which would have been scheduled'
  call say_c '       between midnight and time of invocation.'
  call say_c 'Switch "' || global.eTxtHi || '/R[nn]' || global.eTxtInf || '": check crontab file at least every ''nn'' (default 60) minutes'
  call say_c '       whether it changed; if so, reread it immediately and set the new'
  call say_c '       schedule-times. If given, the crontab file will be checked after dispatching'
  call say_c '       commands. If not given, crontab file will be read only once.'
  call say_c 'Switch "' || global.eTxtHi || '/B' || global.eTxtInf || '": do not colorize output strings (e.g. for usage in more, less'
  call say_c '       etc.). This switch merely suppresses the ANSI-color-sequences attached'
  call say_c '       to the strings (hence BlackAndWhite).'
  call say_c
  call say_c 'examples:' global.eTxtHi || 'CRONRGF /T crontest'
  call say_c '          .. execute statements in file ''crontest'' in testmode'
  call say_c  global.eTxtHi || '          CRONRGF /T /M /R30 crontabfile'
  call say_c '          .. execute statements in the crontab file in testmode, schedule'
  call say_c '             programs which would have been started since midnight and time'
  call say_c '             of invocation, reread crontab file at least every 30 minutes.'
  call say_c  global.eTxtHi || '          CRONRGF /M /R some_file'
  call say_c '          .. execute statements in crontab file ''some_file'', schedule programs which'
  call say_c '             would have been started since midnight and time of invocation,'
  call say_c '             reread ''some_file'' at least every 60 minutes.'
  call say_c  global.eTxtHi || '          CRONRGF /B'
  call say_c '          .. show usage of CRONRGF without colors'

  exit 0

/**
 * write to logfile & close logfile thereafter
 */

Write_log: procedure expose global.

  call lineout global.eLogFile, arg(1)         /* write log-entry */
  call stream global.eLogFile, 'C', 'CLOSE'    /* close logfile */
  return
  /* end Write_log */

/**
  * say with color, if colors enabled
  */

say_c: procedure expose global. stemSchedule.

  SAY global.eTxtInf || arg(1) || global.eScrNorm
  return

/**
 * Format date or date time as ISO
 * @param datetime is yyyymmdd or yyyymmdd hh:mm:ss to format
 * @ if datetime omitted current date/time is formatted
 */

IsoDateTime: procedure
  parse arg ymd hms
  hms = strip(hms)
  if ymd == '' then do
    ymd = date('S')
    hms = time()
  end
  /* Convert yyyymmdd to yyyy-mm-dd
		   string       otbl itbl
		   YYYY-MM-DD         YYYYMMDD */
  ymd = translate('ABCD-EF-GH', ymd, 'ABCDEFGH')
  if hms \== '' then
    ymd = ymd hms
  return ymd
  /* end IsoDateTime */

/*===========================================*/
/*=== SkelRexxFunc compatibility routines ===*/
/*===========================================*/

/*=== LogWriteVTS(message,...) Write_log shim ===*/

LogWriteVTS: procedure expose global.
  s = arg(1)
  say s
  call Write_log IsoDateTime() s
  return

/* end LogWriteVTS */

/*=== LogWriteVTSC(message,...) Write_log shim ===*/

LogWriteVTSC: procedure expose global.
  s = arg(1)
  say s
  call Write_log IsoDateTime() s
  return

/* end LogWriteVTSC */

/*==============================================================================*/
/*=== SkelRexxFunc standards - Delete unused - Move modified above this mark ===*/
/*==============================================================================*/

/*=== CatchError() Catch condition; return gErrCondition ===*/

CatchError:
  gErrCondition = condition('C')
  ErrCondition = gErrCondition		/* 2016-05-30 SHL FIXME to be gone when refs gone */
  return
/* end CatchError */

/*=== ErrorVLX() Set gErrCondition; report to STDOUT and timestamped log file; exit or return if called; never trace ===*/

ErrorVLX:
  parse source . . cmd
  say
  s = sourceline(SIGL)			/* Must cache before call Log... */
  /* Requires LogWriteVTS and dependents */
  /* Requires LogWriteVTSC and dependents */
  gErrCondition = condition('C')
  call LogWriteVTS gErrCondition 'signaled at line' SIGL 'of' cmd'.'
  if condition('D') \= '' then
    call LogWriteVTS condition('D')'.'
  if gErrCondition == 'SYNTAX' & symbol('RC') == 'VAR' then
    call LogWriteVTS 'REXX error =' RC '-' errortext(RC)'.'
  else if symbol('RC') == 'VAR' then
    call LogWriteVTS 'RC =' RC'.'
  call LogWriteVTSC 'Source =' s
  say
  if condition('I') \== 'CALL' | gErrCondition == 'NOVALUE' | gErrCondition == 'SYNTAX' then do
    if symbol('RC') \== 'VAR' then
      RC = 255
    exit RC
  end
  return

/* end ErrorVLX */

/*==========================================================================*/
/*=== SkelRexx standards - Delete unused - Move modified above this mark ===*/
/*==========================================================================*/

